<?php

namespace swf;

/*
    SWF.php: Macromedia Flash (SWF) file parser
    Copyright (C) 2012 Thanos Efraimidis (4real.gr)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/




////////////////////////////////////////////////////////////////////////////////
// Basic IO
////////////////////////////////////////////////////////////////////////////////
use swf2ass\math\RealNumber;

class SWFio {
    public $b; // Byte array (file contents)
    public $bytePos; // Byte position
    public $bitPos; // Bit position

    public function __construct($b) {
        $this->b = $b;
        $this->bytePos = 0;
        $this->bitPos = 0;
    }

    public function doUncompress() {
        $this->b = substr($this->b, 0, 8) . gzuncompress(substr($this->b, 8));
    }

    public function byteAlign() {
        if ($this->bitPos != 0) {
            $this->bytePos++;
            $this->bitPos = 0;
        }
    }

    public function collectBytes($num) {
        $ret = substr($this->b, $this->bytePos, $num);
        $this->bytePos += $num;
        return $ret;
    }

    public function collectBits($num) {
        $value = 0;
        while ($num > 0) {
            $nextbits = ord($this->b[$this->bytePos]);
            $bitsFromHere = 8 - $this->bitPos;
            if ($bitsFromHere > $num) {
                $bitsFromHere = $num;
            }
            $value |= (($nextbits >> (8 - $this->bitPos - $bitsFromHere)) &
                    (0xff >> (8 - $bitsFromHere))) << ($num - $bitsFromHere);
            $num -= $bitsFromHere;
            $this->bitPos += $bitsFromHere;
            if ($this->bitPos >= 8) {
                $this->bitPos = 0;
                $this->bytePos++;
            }
        }
        return $value;
    }

    public function collectString() {
        $beg = $this->bytePos;
        while ($this->collectUI8() != 0) {
            ;
        }
        return substr($this->b, $beg, $this->bytePos - 1 - $beg);
    }

    // Bit values
    public function collectFB($num) : RealNumber { //XXX NOT SURE - NEEDS FIX
        $ret = $this->collectBits($num);
        if (($ret & (1 << ($num - 1))) == 0) {
            // Positive
            $hi = ($ret >> 16) & 0xffff;
            $lo = $ret & 0xffff;
            $ret = (new RealNumber($lo))->divide(65536)->add($hi);
        } else {
            // Negative
            $ret = (1 << $num) - $ret;
            $hi = ($ret >> 16) & 0xffff;
            $lo = $ret & 0xffff;
            $ret = (new RealNumber($lo))->divide(65536)->add($hi)->negate();
        }
        // echo sprintf("collectFB, num is %d, will return [0x%04x]\n", $num, $ret);
        return $ret;
    }

    public function collectUB($num) {
        return $this->collectBits($num);
    }

    public function collectSB($num) {
        $val = $this->collectBits($num);
        if ($num > 0 and ($val & (1 << ($num - 1))) > 0) { // If high bit is set
            $val -= 1 << $num; // Negate
        }
        return $val;
    }

    // Fixed point numbers
    public function collectFixed8() { //TODO: change to RealNumber
        $lo0 = $lo = $this->collectUI8();
        $hi0 = $hi = $this->collectUI8();
        if ($hi < 128) {
            $ret = $hi + $lo / 256.0;
        } else {
            $full = 65536 - (256 * $hi + $lo);
            $hi = $full >> 8;
            $lo = $full & 0xff;
            $ret = -($hi + $lo / 256.0);
        }
        // echo sprintf("collectFixed8 hi=[0x%X], lo=[0x%X], return [%s]\n", $hi0, $lo0, $ret);
        return $ret;
    }

    public function collectFixed() {
        $lo = $this->collectUI16();
        $hi = $this->collectUI16();
        $ret = (new RealNumber($lo))->divide(65536)->add($hi);;
        // echo sprintf("collectFixed hi=[0x%X], lo=[0x%X], return [%s]\n", $hi, $lo, $ret);
        return $ret;
    }

    // Floating point numbers
    public function collectFloat16() {
        $w = $this->collectUI8() << 8;
        $w |= $this->collectUI8();
        $sign = ($w >> 15) & 0x0001; // 1 bit
        $exponent = ($w >> 10) & 0x001f; // 5 bits
        $mantissa = $w & 0x03ff; // 10 bits
        if ($exponent == 0x00) {
            return $sign == 0 ? '0.0' : '-0.0';
        } else if ($exponent == 0x3f) {
            return $mantissa != 0 ? NAN : ($sign == 0 ? INF : -INF);
        } else {
            $ret = $sign == 0 ? 1.0 : -1.0;
            if ($exponent > 16) {
                $ret *= 1 << ($exponent - 16);
            } else if ($exponent < 16) {
                $ret /= 1 << (16 - $exponent);
            }
            $ret *= (1.0 + $mantissa / 1024.0);
            // echo sprintf("float16: w=[0x%X], sign=[%d], exponent=[%d], mantissa=[%d], return=[%s]\n",
            // $w, $sign, $exponent, $mantissa, $ret);
            return $ret;
        }
    }

    public function collectFloat() {
        return unpack("g", $this->collectBytes(4))[1];
    }

    public function collectDouble() {
        return unpack("e", $this->collectBytes(8))[1];
    }

    // Integers
    public function collectUI8() {
        return ord($this->collectBytes(1));
    }

    public function collectSI16() {
        $ret = unpack("v", $this->collectBytes(2))[1];
        if ($ret >= 32768) {
            $ret -= 65536;
        }
        return $ret;
    }

    public function collectUI16() {
        return unpack("v", $this->collectBytes(2))[1];
    }

    public function collectSI32() {
        $ret = unpack("V", $this->collectBytes(4))[1];
        if ($ret > 2147483647) {
            $ret -= 4294967296;
        }
        return $ret;
    }

    public function collectUI32() {
        return unpack("V", $this->collectBytes(4))[1];
    }

    public function collectEncodedU32() {
        $ret = 0;
        $multiplier = 1;
        for (;;) {
            $b = $this->collectUI8();
            $ret += ($b & 0x7f) * $multiplier;
            $multiplier *= 128;
            if (($b & 0x80) == 0) {
                return $ret;
            }
        }
    }

    private static function removeExtraZero($ret) {
        // Keep on removing trailing '00' by '0'
        while (($len = strlen($ret)) > 2 && $ret[$len - 1] == '0' && $ret[$len - 2] == '0') {
            $ret = substr($ret, 0, $len - 1);
        }
        return $ret;
    }

    // For debugging
    public function dumpPosition($len) {
        echo sprintf(" %04d.%1d:", $this->bytePos, $this->bitPos);
        for ($i = 0; $i < $len; $i++) {
            echo sprintf(" 0x%02x", ord($this->b[$this->bytePos + $i]));
        }
        echo sprintf("\n");
    }
}