<?php

namespace swf;

/*
    SWF.php: Macromedia Flash (SWF) file parser
    Copyright (C) 2012 Thanos Efraimidis (4real.gr)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



class SWFrec {
    private SWFio $io; // SWF for basic I/O

    public function __construct(SWFio $io) {
        $this->io = $io;
    }

    public function collectRGB() {
        $ret = array();
        $ret['red'] = $this->io->collectUI8();
        $ret['green'] = $this->io->collectUI8();
        $ret['blue'] = $this->io->collectUI8();
        return $ret;
    }

    public function collectRGBA() {
        $ret = array();
        $ret['red'] = $this->io->collectUI8();
        $ret['green'] = $this->io->collectUI8();
        $ret['blue'] = $this->io->collectUI8();
        $ret['alpha'] = $this->io->collectUI8();
        return $ret;
    }

    public function collectRect() {
        $ret = array();
        $nbits = $this->io->collectUB(5);
        $ret['xmin'] = $this->io->collectSB($nbits);
        $ret['xmax'] = $this->io->collectSB($nbits);
        $ret['ymin'] = $this->io->collectSB($nbits);
        $ret['ymax'] = $this->io->collectSB($nbits);
        $this->io->byteAlign();
        return $ret;
    }

    public function collectMatrix() {
        $ret = array();
        if (($hasScale = $this->io->collectUB(1)) != 0) {
            $nScaleBits = $this->io->collectUB(5);
            $ret['scaleX'] = $this->io->collectFB($nScaleBits);
            $ret['scaleY'] = $this->io->collectFB($nScaleBits);
        }
        if (($hasRotate = $this->io->collectUB(1)) != 0) {
            $nRotateBits = $this->io->collectUB(5);
            $ret['rotateSkew0'] = $this->io->collectFB($nRotateBits);
            $ret['rotateSkew1'] = $this->io->collectFB($nRotateBits);
        }
        if (($nTranslateBits = $this->io->collectUB(5)) != 0) {
            $ret['translateX'] = $this->io->collectSB($nTranslateBits);
            $ret['translateY'] = $this->io->collectSB($nTranslateBits);
        }
        $this->io->byteAlign();
        return $ret;
    }

    public function collectColorTransform($withAlpha) {
        $colorTransform = array();

        $hasAddTerms = $this->io->collectUB(1);
        $hasMultTerms = $this->io->collectUB(1);
        $nbits = $this->io->collectUB(4);
        if ($hasMultTerms != 0) {
            $colorTransform['redMultTerm'] = $this->io->collectSB($nbits);
            $colorTransform['greenMultTerm'] = $this->io->collectSB($nbits);
            $colorTransform['blueMultTerm'] = $this->io->collectSB($nbits);
            if ($withAlpha) {
                $colorTransform['alphaMultTerm'] = $this->io->collectSB($nbits);
            }
        }
        if ($hasAddTerms != 0) {
            $colorTransform['redAddTerm'] = $this->io->collectSB($nbits);
            $colorTransform['greenAddTerm'] = $this->io->collectSB($nbits);
            $colorTransform['blueAddTerm'] = $this->io->collectSB($nbits);
            if ($withAlpha) {
                $colorTransform['alphaAddTerm'] = $this->io->collectSB($nbits);
            }
        }
        $this->io->byteAlign();
        return $colorTransform;
    }

    ////////////////////////////////////////////////////////////////////////////////
    // More complex records
    ////////////////////////////////////////////////////////////////////////////////

    private $actionNames = array
    (// SWF 3
        0x81 => 'ActionGotoFrame',
        0x83 => 'ActionGetURL',
        0x04 => 'ActionNextFrame',
        0x05 => 'ActionPreviousFrame',
        0x06 => 'ActionPlay',
        0x07 => 'ActionStop',
        0x08 => 'ActionToggleQuality',
        0x09 => 'ActionStopSounds',
        0x8a => 'ActionWaitForFrame',
        0x8b => 'ActionSetTarget',
        0x8c => 'ActionGoToLabel',
        // SWF 4
        0x96 => 'ActionPush', // Stack operations
        0x17 => 'ActionPop',
        0x0a => 'ActionAdd', // Arithmetic operators
        0x0b => 'ActionSubtract',
        0x0c => 'ActionMultiply',
        0x0d => 'ActionDivide',
        0x0e => 'ActionEquals', // Numerical comparison
        0x0f => 'ActionLess',
        0x10 => 'ActionAnd', // Logical operators
        0x11 => 'ActionOr',
        0x12 => 'ActionNot',
        0x13 => 'ActionStringEquals', // String manipulation
        0x14 => 'ActionStringLength',
        0x21 => 'ActionStringAdd',
        0x15 => 'ActionStringExtract',
        0x29 => 'ActionStringLess',
        0x31 => 'ActionMBStringLength',
        0x35 => 'ActionMBStringExtract',
        0x18 => 'ActionToInteger', // Type conversion
        0x32 => 'ActionCharToAscii',
        0x33 => 'ActionAsciiToChar',
        0x36 => 'ActionMBCharToAscii',
        0x37 => 'ActionMBAsciiToChar',
        0x99 => 'ActionJump', // Control flow
        0x9d => 'ActionIf',
        0x9e => 'ActionCall',
        0x1c => 'ActionGetVariable', // Variables
        0x1d => 'ActionSetVariable',
        0x9a => 'ActionGetURL2', // Movie control
        0x9f => 'ActionGotoFrame2',
        0x20 => 'ActionSetTarget2',
        0x22 => 'ActionGetProperty',
        0x23 => 'ActionSetProperty',
        0x24 => 'ActionCloneSprite',
        0x25 => 'ActionRemoteSprite',
        0x27 => 'ActionStartDrag',
        0x28 => 'ActionEndDrag',
        0x8d => 'ActionWaitForFrame2',
        0x26 => 'ActionTrace', // Utilities
        0x34 => 'ActionGetTime',
        0x30 => 'ActionRandomNumber',
        // SWF 5
        0x3d => 'ActionCallFunction', // ScriptObject actions
        0x52 => 'ActionCallMethod',
        0x88 => 'ActionConstantPool',
        0x9b => 'ActionDefineFunction',
        0x3c => 'ActionDefineLocal',
        0x41 => 'ActionDefineLocal2',
        0x3a => 'ActionDelete',
        0x3b => 'ActionDelete2',
        0x46 => 'ActionEnumerate',
        0x49 => 'ActionEquals2',
        0x4e => 'ActionGetMember',
        0x42 => 'ActionInitArray',
        0x43 => 'ActionInitiObject',
        0x53 => 'ActionNewMethod',
        0x40 => 'ActionNewObject',
        0x4f => 'ActionSetMember',
        0x45 => 'ActionTargetPath',
        0x94 => 'ActionWith',
        0x4a => 'ActionToNumber', // Type actions
        0x4b => 'ActionToString',
        0x44 => 'ActionTypeOf',
        0x47 => 'ActionAdd2', // Math actions
        0x48 => 'ActionLess2',
        0x3f => 'ActionModule',
        0x60 => 'ActionBitAnd', // Stack operator actions
        0x63 => 'ActionBitLShift',
        0x61 => 'ActionBitOr',
        0x64 => 'ActionBitRShift',
        0x65 => 'ActionBitURShift',
        0x62 => 'ActionBitXor',
        0x51 => 'ActionDecrement',
        0x50 => 'ActionIncrement',
        0x4c => 'ActionPushDuplicate',
        0x3e => 'ActionReturn',
        0x4d => 'ActionStackSwap',
        0x87 => 'ActionStoreRegister',
        // SWF 6
        0x59 => 'DoInitAction',
        0x54 => 'ActionInstanceOf',
        0x55 => 'ActionEnumerate2',
        0x66 => 'ActionStrictEquals',
        0x67 => 'ActionGreater',
        0x68 => 'ActionStringGreater',
        // SWF 7
        0x8e => 'ActionDefineFunction2',
        0x69 => 'ActionExtends',
        0x2b => 'ActionCastOp',
        0x2c => 'ActionImplementsOp',
        0x8f => 'ActionTry',
        0x2a => 'ActionThrow',
        // SWF 9
        0x82 => 'DoABC',
        // SWF 10
    );

    private function actionRecordIsProcessed($offset, $actions) {
        foreach ($actions as $action) {
            if ($action['offset'] == $offset) {
                return true;
            }
        }
        return false;
    }

    public function collectActionRecords($bytePosEnd) {
        $actions = array();
        for (;;) {
            if ($this->io->bytePos >= $bytePosEnd) {
                break;
            }
            $offset = $this->io->bytePos;
            $actionLength = 0;
            $actionName = null;
            $actionData = null;
            if (($actionCode = $this->io->collectUI8()) == 0) {
                // echo sprintf("%6d: Code=0x%02x, breaking\n", $offset, $actionCode);
                $actions[] =
                    array('offset' => $offset,
                        'actionCode' => $actionCode,
                        'actionLength' => 0,
                        'actionName' => null,
                        'actionData' => null);
                continue; // break;
            }
            if ($actionCode >= 0x80) {
                $actionLength = $this->io->collectUI16();
            }
            if (isset($this->actionNames[$actionCode])) {
                $actionName = $this->actionNames[$actionCode];
                if ($actionLength > 0) {
                    $actionData = $this->collectActionData($actionCode, $actionLength);
                }
                // echo sprintf("%6d: Code=0x%02x, length=%d, name=%s\n", $offset, $actionCode, $actionLength, $actionName);
                $actions[] =
                    array('offset' => $offset,
                        'actionCode' => $actionCode,
                        'actionLength' => $actionLength,
                        'actionName' => $actionName,
                        'actionData' => $actionData);
            } else {
                throw new \Exception(sprintf("Internal error: actionCode=0x%02X, actionLength=%d", $actionCode, $actionLength));
            }
        }
        if ($this->io->bytePos != $bytePosEnd) {
            echo sprintf("\n\nTHANOS THERE ARE %d bytes left\n", $bytePosEnd - $this->io->bytePos);
            // throw new \Exception("OOPS");
        }
        // var_dump($actions);
        return $actions;
    }

    public function collectActionData($actionCode, $actionLength) {
        $actionData = array();
        if ($actionCode == 0x81) { // ActionGotoFrame
            $actionData['frame'] = $this->io->collectUI16();
        } else if ($actionCode == 0x83) { // ActionGetURL
            $actionData['url'] = $this->io->collectString();
            $actionData['target'] = $this->io->collectString();
        } else if ($actionCode == 0x87) { // ActionStoreRegister
            $actionData['registerNumber'] = $this->io->collectUI8();
        } else if ($actionCode == 0x88) { // ActionConstantPool
            $count = $this->io->collectUI16();
            for ($i = 0; $i < $count; $i++) {
                $actionData[sprintf('constant_%d', $i)] = $this->io->collectString();
            }
        } else if ($actionCode == 0x8A) { // ActionWaitForFrame
            $actionData['frame'] = $this->io->collectUI16();
            $actionData['skipCount'] = $this->io->collectUI8();
        } else if ($actionCode == 0x8B) { // ActionSetTarget
            $actionData['targetName'] = $this->io->collectString();
        } else if ($actionCode == 0x8C) { // ActionGoToLabel
            $actionData['label'] = $this->io->collectString();
        } else if ($actionCode == 0x8D) { // ActionWaitForFrame2
            $actionData['skipCount'] = $this->io->collectUI8();
        } else if ($actionCode == 0x8E) { // ActionDefineFunction2
            $actionData['functionName'] = $this->io->collectString();
            $numParams = $this->io->collectUI16();
            $actionData['registerCount'] = $this->io->collectUI8();

            $actionData['preloadParentFlag'] = $this->io->collectUB(1);
            $actionData['preloadRootFlag'] = $this->io->collectUB(1);
            $actionData['suppressSuperFlag'] = $this->io->collectUB(1);
            $actionData['preloadSuperFlag'] = $this->io->collectUB(1);
            $actionData['suppressArgumentsFlag'] = $this->io->collectUB(1);
            $actionData['preloadArgumentsFlag'] = $this->io->collectUB(1);
            $actionData['suppressThisFlag'] = $this->io->collectUB(1);
            $actionData['preloadThisFlag'] = $this->io->collectUB(1);

            $this->io->collectUB(7); // Reserved
            $actionData['preloadGlobalFlag'] = $this->io->collectUB(1);

            for ($i = 0; $i < $numParams; $i++) {
                $actionData[sprintf('register_%d', $i)] = $this->io->collectUI8();
                $actionData[sprintf('param_%d', $i)] = $this->io->collectString();
            }
            $actionData['codeSize'] = $this->io->collectUI16();
        } else if ($actionCode == 0x94) { // ActionWith
            $size = $this->io->collectUI16();
            $actionData['code'] = $this->io->collectBytes($size);
        } else if ($actionCode == 0x96) { // ActionPush
            $bytePosEnd = $this->io->bytePos + $actionLength;
            while ($this->io->bytePos < $bytePosEnd) {
                $type = $this->io->collectUI8();
                switch ($type) {
                    case 0: // String
                        $val = $this->io->collectString();
                        break;
                    case 1: // Float
                        $val = $this->io->collectFloat();
                        break;
                    case 2: // Null
                        $val = null;
                        break;
                    case 3: // Undefined
                        $val = null;
                        break;
                    case 4: // Register
                        $val = $this->io->collectUI8();
                        break;
                    case 5: // Boolean
                        $val = $this->io->collectUI8();
                        break;
                    case 6: // Double
                        $val = $this->io->collectDouble();
                        break;
                    case 7: // Integer
                        $val = $this->io->collectUI32();
                        break;
                    case 8: // Constant8
                        $val = $this->io->collectUI8();
                        break;
                    case 9: // Constant16
                        $val = $this->io->collectUI16();
                        break;
                    default:
                        throw new \Exception(sprintf("Internal error: type=%d", $type));
                }
                $actionData[] = array('type' => $type, 'val' => $val);
            }
        } else if ($actionCode == 0x99 || $actionCode == 0x9D) { // ActionJump or ActionIf
            $actionData['branchOffset'] = $this->io->collectSI16();
        } else if ($actionCode == 0x9A) { // ActionGetURL2
            $actionData['sendVarsMethod'] = $this->io->collectUB(2);
            $this->io->collectUB(4); // Reserved
            $actionData['loadTargetFlag'] = $this->io->collectUB(1);
            $actionData['loadVariablesFlag'] = $this->io->collectUB(1);
        } else if ($actionCode == 0x9B) { // ActionDefineFunction
            $actionData['functionName'] = $this->io->collectString();
            $numParams = $this->io->collectUI16();
            for ($i = 0; $i < $numParams; $i++) {
                $actionData[sprintf('param_%d', $i)] = $this->io->collectString();
            }
            $actionData['codeSize'] = $this->io->collectUI16();
        } else if ($actionCode == 0x9F) { // ActionGotoFrame2
            $this->io->collectUB(6); // Reserved
            $actionData['sceneBiasFlag'] = $this->io->collectUB(1);
            $actionData['playFlag'] = $this->io->collectUB(1);
            if ($actionData['sceneBiasFlag'] == 1) {
                $actionData['sceneBias'] = $this->io->collectUI16();
            }
        } else {
            throw new \Exception(sprintf("Internal error: actionCode=0x%02X, actionLength=%d", $actionCode, $actionLength));
        }
        return $actionData;
    }

    public function collectShape($shapeVersion) {
        $numFillBits = $this->io->collectUB(4);
        $numLineBits = $this->io->collectUB(4);
        $shapeRecords = $this->collectShapeRecords($shapeVersion, null, null, $numFillBits, $numLineBits);
        return $shapeRecords;
    }

    public function collectShapeWithStyle($shapeVersion) {
        $ret = array();
        $ret['fillStyles'] = $this->collectFillStyleArray($shapeVersion);
        $ret['lineStyles'] = $this->collectLineStyleArray($shapeVersion);
        $numFillBits = $this->io->collectUB(4);
        $numLineBits = $this->io->collectUB(4);
        $ret['shapeRecords'] = $this->collectShapeRecords($shapeVersion, $ret['fillStyles'], $ret['lineStyles'], $numFillBits, $numLineBits);
        return $ret;
    }

    public function collectShapeRecords($shapeVersion, $fillStyles, $lineStyles, $numFillBits, $numLineBits) {
        $shapeRecords = array();
        for (;;) {
            $typeFlag = $this->io->collectUB(1);
            if ($typeFlag == 0) {
                $stateNewStyles = $this->io->collectUB(1);
                $stateLineStyle = $this->io->collectUB(1);
                $stateFillStyle1 = $this->io->collectUB(1);
                $stateFillStyle0 = $this->io->collectUB(1);
                $stateMoveTo = $this->io->collectUB(1);
                if ($stateNewStyles == 0 && $stateLineStyle == 0 && $stateFillStyle1 == 0 && $stateFillStyle0 == 0 && $stateMoveTo == 0) {
                    // EndShapeRecord
                    $shapeRecords[] = array('type' => 'EndShapeRecord');
                    break;
                } else {
                    // StyleChangeRecord
                    $shapeRecord = array();
                    $shapeRecord['type'] = 'StyleChangeRecord';
                    $shapeRecord['stateNewStyles'] = $stateNewStyles;
                    $shapeRecord['stateLineStyle'] = $stateLineStyle;
                    $shapeRecord['stateFillStyle1'] = $stateFillStyle1;
                    $shapeRecord['stateFillStyle0'] = $stateFillStyle0;
                    $shapeRecord['stateMoveTo'] = $stateMoveTo;

                    if ($shapeRecord['stateMoveTo'] != 0) {
                        $moveBits = $this->io->collectUB(5);
                        $shapeRecord['moveDeltaX'] = $this->io->collectSB($moveBits);
                        $shapeRecord['moveDeltaY'] = $this->io->collectSB($moveBits);
                    }
                    if ($shapeRecord['stateFillStyle0'] != 0) {
                        $shapeRecord['fillStyle0'] = $this->io->collectUB($numFillBits);
                    }
                    if ($shapeRecord['stateFillStyle1'] != 0) {
                        $shapeRecord['fillStyle1'] = $this->io->collectUB($numFillBits);
                    }
                    if ($shapeRecord['stateLineStyle'] != 0) {
                        $shapeRecord['lineStyle'] = $this->io->collectUB($numLineBits);
                    }
                    if ($shapeRecord['stateNewStyles'] != 0 && ($shapeVersion == 2 || $shapeVersion == 3 || $shapeVersion == 4)) { // XXX shapeVersion 4 not in spec
                        $this->io->byteAlign();
                        $shapeRecord['fillStyles'] = $fillStyles = $this->collectFillStyleArray($shapeVersion);
                        $shapeRecord['lineStyles'] = $lineStyles = $this->collectLineStyleArray($shapeVersion);
                        $numFillBits = $this->io->collectUB(4);
                        $numLineBits = $this->io->collectUB(4);
                    }
                    $shapeRecords[] = $shapeRecord;
                }
            } else {
                $straightFlag = $this->io->collectUB(1);
                if ($straightFlag == 1) {
                    // StraightEdgeRecord
                    $shapeRecord = array();
                    $shapeRecord['type'] = 'StraightEdgeRecord';
                    $numBits = $this->io->collectUB(4);
                    $shapeRecord['generalLineFlag'] = $this->io->collectUB(1);
                    if ($shapeRecord['generalLineFlag'] == 0) {
                        $shapeRecord['vertLineFlag'] = $this->io->collectUB(1);
                    }
                    if ($shapeRecord['generalLineFlag'] == 1 || $shapeRecord['vertLineFlag'] == 0) {
                        $shapeRecord['deltaX'] = $this->io->collectSB($numBits + 2);
                    }
                    if ($shapeRecord['generalLineFlag'] == 1 || $shapeRecord['vertLineFlag'] == 1) {
                        $shapeRecord['deltaY'] = $this->io->collectSB($numBits + 2);
                    }
                    $shapeRecords[] = $shapeRecord;
                } else {
                    // CurvedEdgeRecord
                    $shapeRecord = array();
                    $shapeRecord['type'] = 'CurvedEdgeRecord';
                    $numBits = $this->io->collectUB(4);
                    $shapeRecord['controlDeltaX'] = $this->io->collectSB($numBits + 2);
                    $shapeRecord['controlDeltaY'] = $this->io->collectSB($numBits + 2);
                    $shapeRecord['anchorDeltaX'] = $this->io->collectSB($numBits + 2);
                    $shapeRecord['anchorDeltaY'] = $this->io->collectSB($numBits + 2);
                    $shapeRecords[] = $shapeRecord;
                }
            }
        }
        $this->io->byteAlign();
        return $shapeRecords;
    }

    public function collectMorphFillStyleArray() {
        $morphFillStyleArray = array();
        $fillStyleCount = $this->io->collectUI8();
        if ($fillStyleCount == 0xff) {
            $fillStyleCount = $this->io->collectUI16(); // Extended
        }
        for ($i = 0; $i < $fillStyleCount; $i++) {
            $morphFillStyleArray[] = $this->collectMorphFillStyle();
        }
        return $morphFillStyleArray;
    }

    public function collectMorphFillStyle() {
        $morphFillStyle = array(); // To return
        $morphFillStyle['fillStyleType'] = $this->io->collectUI8();
        switch($morphFillStyle['fillStyleType']) {
            case 0x00: // Solid fill
                $morphFillStyle['startColor'] = $this->collectRGBA();
                $morphFillStyle['endColor'] = $this->collectRGBA();
                break;
            case 0x10: // Linear gradient fill
            case 0x12: // Radial gradient fill
                $morphFillStyle['startGradientMatrix'] = $this->collectMatrix();
                $morphFillStyle['endGradientMatrix'] = $this->collectMatrix();
                $morphFillStyle['gradient'] = $this->collectMorphGradient();
                break;
            case 0x40: // Repeating bitmap
            case 0x41: // Clipped bitmap fill
            case 0x42: // Non-smoothed repeating bitmap
            case 0x43: // Non-smoothed clipped bitmap
                $morphFillStyle['bitmapId'] = $this->io->collectUI16();
                $morphFillStyle['startBitmapMatrix'] = $this->collectMatrix();
                $morphFillStyle['endBitmapMatrix'] = $this->collectMatrix();
                break;
            default:
                throw new \Exception(sprintf('Internal error: fillStyleType=%d', $morphFillStyle['fillStyleType']));
        }
        return $morphFillStyle;
    }

    public function collectMorphGradient() {
        $morphGradient = array();
        $numGradients = $this->io->collectUI8();
        for ($i = 0; $i < $numGradients; $i++) {
            $morphGradient[] = $this->collectMorphGradientRecord();
        }
        return $morphGradient;
    }

    public function collectMorphGradientRecord() {
        $morphGradientRecord = array();
        $morphGradientRecord['startRatio'] = $this->io->collectUI8();
        $morphGradientRecord['startColor'] = $this->collectRGBA();
        $morphGradientRecord['endRatio'] = $this->io->collectUI8();
        $morphGradientRecord['endColor'] = $this->collectRGBA();
        return $morphGradientRecord;
    }

    public function collectMorphLineStyleArray($version) {
        $morphLineStyleArray = array();
        $lineStyleCount = $this->io->collectUI8();
        if ($lineStyleCount == 0xff) {
            $lineStyleCount = $this->io->collectUI16();
        }
        if ($version == 1) {
            for ($i = 0; $i < $lineStyleCount; $i++) {
                $morphLineStyleArray[] = $this->collectMorphLineStyle();
            }
        } else if ($version == 2) {
            for ($i = 0; $i < $lineStyleCount; $i++) {
                $morphLineStyleArray[] = $this->collectMorphLineStyle2();
            }
        } else {
            throw new \Exception(sprintf('Internal error: version=%d', $version));
        }
        return $morphLineStyleArray;
    }

    public function collectMorphLineStyle() {
        $morphLineStyle = array(); // To return
        $morphLineStyle['startWidth'] = $this->io->collectUI16();
        $morphLineStyle['endWidth'] = $this->io->collectUI16();
        $morphLineStyle['startColor'] = $this->collectRGBA();
        $morphLineStyle['endColor'] = $this->collectRGBA();
        return $morphLineStyle;
    }

    public function collectMorphLineStyle2() {
        $morphLineStyle2 = array(); // To return
        $morphLineStyle2['startWidth'] = $this->io->collectUI16();
        $morphLineStyle2['endWidth'] = $this->io->collectUI16();

        $morphLineStyle2['startCapStyle'] = $this->io->collectUB(2);
        $morphLineStyle2['joinStyle'] = $this->io->collectUB(2);
        $morphLineStyle2['hasFillFlag'] = $this->io->collectUB(1);
        $morphLineStyle2['noHScaleFlag'] = $this->io->collectUB(1);
        $morphLineStyle2['noVScaleFlag'] = $this->io->collectUB(1);
        $morphLineStyle2['pixelHintingFlag'] = $this->io->collectUB(1);

        $this->io->collectUB(5); // Reserved
        $morphLineStyle2['noClose'] = $this->io->collectUB(1);
        $morphLineStyle2['endCapStyle'] = $this->io->collectUB(2);

        if ($morphLineStyle2['joinStyle'] == 2) {
            $morphLineStyle2['miterLimitFactor'] = $this->io->collectUI16();
        }
        if ($morphLineStyle2['hasFillFlag'] == 0) {
            $morphLineStyle2['startColor'] = $this->collectRGBA();
            $morphLineStyle2['endColor'] = $this->collectRGBA();
        }
        if ($morphLineStyle2['hasFillFlag'] == 1) {
            $morphLineStyle2['fillType'] = $this->collectMorphFillStyle();
        }
        return $morphLineStyle2;
    }

    public function collectGradient($shapeVersion) {
        $gradient = array();
        $gradient['spreadMode'] = $this->io->collectUB(2);
        $gradient['interpolationMode'] = $this->io->collectUB(2);
        $numGradientRecords = $this->io->collectUB(4);
        $gradient['gradientRecords'] = $this->collectGradientRecords($numGradientRecords, $shapeVersion);
        return $gradient;
    }

    public function collectFocalGradient($shapeVersion) { // shapeVersion must be 4
        $focalGradient = array();
        $focalGradient['spreadMode'] = $this->io->collectUB(2);
        $focalGradient['interpolationMode'] = $this->io->collectUB(2);
        $numGradientRecords = $this->io->collectUB(4);
        $focalGradient['gradientRecords'] = $this->collectGradientRecords($numGradientRecords, $shapeVersion);
        $focalGradient['focalPoint'] = $this->io->collectFixed8();
        return $focalGradient;
    }

    public function collectFilterList() {
        $filterList = array();
        $numberOfFilters = $this->io->collectUI8();
        for ($f = 0; $f < $numberOfFilters; $f++) {
            $filter = array();
            $filter['filterId'] = $this->io->collectUI8();
            switch ($filter['filterId']) {
                case 0: // DropShadowFilter
                    $filter['dropShadowColor'] = $this->collectRGBA();
                    $filter['blurX'] = $this->io->collectFixed();
                    $filter['blurY'] = $this->io->collectFixed();
                    $filter['angle'] = $this->io->collectFixed();
                    $filter['distance'] = $this->io->collectFixed();
                    $filter['strength'] = $this->io->collectFixed8();
                    $filter['innerShadow'] = $this->io->collectUB(1);
                    $filter['knockout'] = $this->io->collectUB(1);
                    $filter['compositeSource'] = $this->io->collectUB(1);
                    $filter['passes'] = $this->io->collectUB(5);
                    break;
                case 1: // BlurFilter
                    $filter['blurX'] = $this->io->collectFixed();
                    $filter['blurY'] = $this->io->collectFixed();
                    $filter['passes'] = $this->io->collectUB(5);
                    $this->io->collectUB(3); // Reserved, must be 0
                    break;
                case 2: // GlowFilter
                    $filter['glowColor'] = $this->collectRGBA();
                    $filter['blurX'] = $this->io->collectFixed();
                    $filter['blurY'] = $this->io->collectFixed();
                    $filter['strength'] = $this->io->collectFixed8();
                    $filter['innerGlow'] = $this->io->collectUB(1);
                    $filter['knockout'] = $this->io->collectUB(1);
                    $filter['compositeSource'] = $this->io->collectUB(1);
                    $filter['passes'] = $this->io->collectUB(5);
                    break;
                case 3: // BevelFilter
                    $filter['hadowColor'] = $this->collectRGBA();
                    $filter['highlightColor'] = $this->collectRGBA();
                    $filter['blurX'] = $this->io->collectFixed();
                    $filter['blurY'] = $this->io->collectFixed();
                    $filter['angle'] = $this->io->collectFixed();
                    $filter['distance'] = $this->io->collectFixed();
                    $filter['strength'] = $this->io->collectFixed8();
                    $filter['innerShadow'] = $this->io->collectUB(1);
                    $filter['knockout'] = $this->io->collectUB(1);
                    $filter['compositeSource'] = $this->io->collectUB(1);
                    $filter['onTop'] = $this->io->collectUB(1);
                    $filter['passes'] = $this->io->collectUB(4);
                    break;
                case 4: // GradientGlowFilter
                    $filter['numColors'] = $this->io->collectUI8();
                    $filter['gradientColors'] = array();
                    for ($i = 0; $i < $filter['numColors']; $i++) {
                        $filter['gradientColors'][] = $this->collectRGBA();
                    }
                    $filter['gradientRatio'] = array();
                    for ($i = 0; $i < $filter['numColors']; $i++) {
                        $filter['gradientRatio'][] = $this->io->collectUI8();
                    }
                    $filter['blurX'] = $this->io->collectFixed();
                    $filter['blurY'] = $this->io->collectFixed();
                    $filter['angle'] = $this->io->collectFixed();
                    $filter['distance'] = $this->io->collectFixed();
                    $filter['strength'] = $this->io->collectFixed8();
                    $filter['innerShadow'] = $this->io->collectUB(1);
                    $filter['knockout'] = $this->io->collectUB(1);
                    $filter['compositeSource'] = $this->io->collectUB(1);
                    $filter['onTop'] = $this->io->collectUB(1);
                    $filter['passes'] = $this->io->collectUB(4);
                    break;
                case 5: // ConvolutionFilter
                    $filter['matrixX'] = $this->io->collectUI8();
                    $filter['matrixY'] = $this->io->collectUI8();
                    $filter['divisor'] = $this->collectFloat();
                    $filter['bias'] = $this->collectFloat();
                    $filter['matrix'] = array();
                    for ($i = 0; $i < $filter['matrixX'] * $filter['matrixY']; $i++) {
                        $filter['matrix'][] = $this->collectFloat();
                    }
                    $filter['defaultColor'] = $this->collectRGBA();
                    $this->io->collectUB(6);
                    $filter['clamp'] = $this->io->collectUB(1);
                    $filter['preservedAlpha'] = $this->io->collectUB(1);
                    break;
                case 6: // ColorMatrixFilter
                    $matrix = array();
                    for ($i = 0; $i < 20; $i++) {
                        $matrix[$i] = $this->io->collectFloat();
                    }
                    $filter['matrix'] = $matrix;
                    break;
                case 7: // GradientBevelFilter
                    $filter['numColors'] = $this->io->collectUI8();
                    $filter['gradientColors'] = array();
                    for ($i = 0; $i < $filter['numColors']; $i++) {
                        $filter['gradientColors'][] = $this->collectRGBA();
                    }
                    $filter['gradientRatio'] = array();
                    for ($i = 0; $i < $filter['numColors']; $i++) {
                        $filter['gradientRatio'][] = $this->io->collectUI8();
                    }
                    $filter['blurX'] = $this->io->collectFixed();
                    $filter['blurY'] = $this->io->collectFixed();
                    $filter['angle'] = $this->io->collectFixed();
                    $filter['distance'] = $this->io->collectFixed();
                    $filter['strength'] = $this->io->collectFixed8();
                    $filter['innerShadow'] = $this->io->collectUB(1);
                    $filter['knockout'] = $this->io->collectUB(1);
                    $filter['compositeSource'] = $this->io->collectUB(1);
                    $filter['onTop'] = $this->io->collectUB(1);
                    $filter['passes'] = $this->io->collectUB(4);
                    break;
                default:
                    throw new \Exception(sprintf('Internal error: filterId=%d', $filter['filterId']));
            }
            $filterList[] = $filter;
        }
        return $filterList;
    }

    public function collectSoundInfo() {
        $soundInfo = array();

        $this->io->collectUB(2); // Reserved
        $soundInfo['syncStop'] = $this->io->collectUB(1);
        $soundInfo['syncNoMultiple'] = $this->io->collectUB(1);
        $soundInfo['hasEnvelope'] = $this->io->collectUB(1);
        $soundInfo['hasLoops'] = $this->io->collectUB(1);
        $soundInfo['hasOutPoint'] = $this->io->collectUB(1);
        $soundInfo['hasInPoint'] = $this->io->collectUB(1);

        if ($soundInfo['hasInPoint'] != 0) {
            $soundInfo['inPoint'] = $this->io->collectUI32();
        }
        if ($soundInfo['hasOutPoint'] != 0) {
            $soundInfo['outPoint'] = $this->io->collectUI32();
        }
        if ($soundInfo['hasLoops'] != 0) {
            $soundInfo['loopCount'] = $this->io->collectUI16();
        }
        if ($soundInfo['hasEnvelope'] != 0) {
            $soundInfo['envelopeRecords'] = array();
            $envPoints = $this->io->collectUI8();
            for ($i = 0; $i < $envPoints; $i++) {
                $soundEnvelope = array();
                $soundEnvelope['pos44'] = $this->io->collectUI32();
                $soundEnvelope['leftLevel'] = $this->io->collectUI16();
                $soundEnvelope['rightLevel'] = $this->io->collectUI16();
                $soundInfo['envelopeRecords'][] = $soundEnvelope;
            }
        }
        return $soundInfo;
    }

    public function collectButtonRecords($version) {
        $buttonRecords = array();
        for (;;) {
            $buttonRecord = array();

            $reserved = $this->io->collectUB(2);
            $buttonRecord['buttonHasBlendMode'] = $this->io->collectUB(1);
            $buttonRecord['buttonHasFilterList'] = $this->io->collectUB(1);
            $buttonRecord['buttonStateHitTest'] = $this->io->collectUB(1);
            $buttonRecord['buttonStateDown'] = $this->io->collectUB(1);
            $buttonRecord['buttonStateOver'] = $this->io->collectUB(1);
            $buttonRecord['buttonStateUp'] = $this->io->collectUB(1);

            if ($reserved == 0 &&
                $buttonRecord['buttonHasBlendMode'] == 0 &&
                $buttonRecord['buttonHasFilterList'] == 0 &&
                $buttonRecord['buttonStateHitTest'] == 0 &&
                $buttonRecord['buttonStateDown'] == 0 &&
                $buttonRecord['buttonStateOver'] == 0 &&
                $buttonRecord['buttonStateUp'] == 0) {
                break;
            }

            $buttonRecord['characterId'] = $this->io->collectUI16();
            $buttonRecord['placeDepth'] = $this->io->collectUI16();
            $buttonRecord['placeMatrix'] = $this->collectMatrix();
            if ($version == 2) {
                $buttonRecord['colorTransform'] = $this->collectColorTransform(true);
            }
            if ($version == 2 && $buttonRecord['buttonHasFilterList'] != 0) {
                $buttonRecord['filterList'] = $this->collectFilterList();
            }
            if ($version == 2 && $buttonRecord['buttonHasBlendMode'] != 0) {
                $buttonRecord['blendMode'] = $this->io->collectUI8();
            }
            $buttonRecords[] = $buttonRecord;
        }
        return $buttonRecords;
    }

    public function collectButtonCondActions($bytePosEnd) {
        $buttonCondActions = array();
        for (;;) {
            $buttonCondAction = array();
            $here = $this->io->bytePos;
            $condActionSize = $this->io->collectUI16();

            $buttonCondAction['condIdleToOverDown'] = $this->io->collectUB(1);
            $buttonCondAction['condOutDownToIdle'] = $this->io->collectUB(1);
            $buttonCondAction['condOutDownToOverDown'] = $this->io->collectUB(1);
            $buttonCondAction['condOverDownToOutDown'] = $this->io->collectUB(1);
            $buttonCondAction['condOverDownToOverUp'] = $this->io->collectUB(1);
            $buttonCondAction['condOverUpToOverDown'] = $this->io->collectUB(1);
            $buttonCondAction['condOverUpToIdle'] = $this->io->collectUB(1);
            $buttonCondAction['condIdleToOverUp'] = $this->io->collectUB(1);

            $buttonCondAction['condKeyPress'] = $this->io->collectUB(7);
            $buttonCondAction['condOverDownToIdle'] = $this->io->collectUB(1);

            $buttonCondAction['actions'] = $this->collectActionRecords($condActionSize == 0 ? $bytePosEnd : $here + $condActionSize);

            $buttonCondActions[] = $buttonCondAction;
            if ($condActionSize == 0) {
                break;
            }
        }
        return $buttonCondActions;
    }

    public function collectClipActions($swfVersion) {
        $clipActions = array();
        $this->io->collectUI16(); // Reserved, must be 0
        $clipActions['allEventFlags'] = $this->collectClipEventFlags($swfVersion);
        $clipActions['clipActionRecords'] = array();
        for (;;) {
            // Collect clipActionEndFlag, if zero then break, if not zero then push back
            if ($swfVersion <= 5) {
                if (($endFlag = $this->io->collectUI16()) == 0) {
                    break;
                }
                $this->io->bytePos -= 2;
            } else {
                if (($endFlag = $this->io->collectUI32()) == 0) {
                    break;
                }
                $this->io->bytePos -= 4;
            }
            $clipActions['clipActionRecords'][] = $this->collectClipActionRecord($swfVersion);
        }
        return $clipActions;
    }

    public function collectClipActionRecord($swfVersion) {
        $clipActionRecord = array();
        $clipActionRecord['eventFlags'] = $this->collectClipEventFlags($swfVersion);
        $actionRecordSize = $this->io->collectUI32();
        $here = $this->io->bytePos;
        if (isset($clipActionRecord['eventFlags']['clipEventKeyPress']) && $clipActionRecord['eventFlags']['clipEventKeyPress'] == 1) {
            $clipActionRecord['keyCode'] = $this->io->collectUI8();
        }
        $clipActionRecord['actions'] = $this->collectActionRecords($here + $actionRecordSize);
        return $clipActionRecord;
    }

    public function collectClipEventFlags($swfVersion) {
        $ret = array();
        $ret['clipEventKeyUp'] = $this->io->collectUB(1);
        $ret['clipEventKeyDown'] = $this->io->collectUB(1);
        $ret['clipEventMouseUp'] = $this->io->collectUB(1);
        $ret['clipEventMouseDown'] = $this->io->collectUB(1);
        $ret['clipEventMouseMove'] = $this->io->collectUB(1);
        $ret['clipEventUnload'] = $this->io->collectUB(1);
        $ret['clipEventEnterFrame'] = $this->io->collectUB(1);
        $ret['clipEventLoad'] = $this->io->collectUB(1);

        $ret['clipEventDragOver'] = $this->io->collectUB(1);
        $ret['clipEventRollOut'] = $this->io->collectUB(1);
        $ret['clipEventRollOver'] = $this->io->collectUB(1);
        $ret['clipEventReleaseOutside'] = $this->io->collectUB(1);
        $ret['clipEventRelease'] = $this->io->collectUB(1);
        $ret['clipEventPress'] = $this->io->collectUB(1);
        $ret['clipEventInitialize'] = $this->io->collectUB(1);
        $ret['clipEventData'] = $this->io->collectUB(1);

        if ($swfVersion >= 6) {
            $this->io->collectUB(5); // Reserved
            $ret['clipEventConstruct'] = $this->io->collectUB(1);
            $ret['clipEventKeyPress'] = $this->io->collectUB(1);
            $ret['clipEventDragOut'] = $this->io->collectUB(1);
            $this->io->collectUB(8); // Reserved
        }
        return $ret;
    }

    public function collectGradientRecords($numGradientRecords, $shapeVersion) {
        $gradientRecords = array();
        for ($i = 0; $i < $numGradientRecords; $i++) {
            $gradientRecord = array();
            $gradientRecord['ratio'] = $this->io->collectUI8();
            if ($shapeVersion == 1 || $shapeVersion == 2) {
                $gradientRecord['color'] = $this->collectRGB();
            } else if ($shapeVersion == 3 || $shapeVersion == 4) { //XXX shapeVersion 4 not in spec
                $gradientRecord['color'] = $this->collectRGBA();
            } else {
                throw new \Exception(sprintf('Internal error: shapeVersion=%d', $shapeVersion));
            }
            $gradientRecords[] = $gradientRecord;
        }
        return $gradientRecords;
    }

    public function collectTextRecords($glyphBits, $advanceBits, $textVersion) {
        $textRecords = array();
        // Collect text records
        for (;;) {
            $textRecord = array();
            $textRecord['textRecordType'] = $this->io->collectUB(1);
            $reserved = $this->io->collectUB(3); // Reserved, must be 0
            $textRecord['styleFlagsHasFont'] = $this->io->collectUB(1);
            $textRecord['styleFlagsHasColor'] = $this->io->collectUB(1);
            $textRecord['styleFlagsHasYOffset'] = $this->io->collectUB(1);
            $textRecord['styleFlagsHasXOffset'] = $this->io->collectUB(1);

            if ($textRecord['textRecordType'] == 0 &&
                $textRecord['styleFlagsHasFont'] == 0 && $textRecord['styleFlagsHasColor'] == 0 &&
                $textRecord['styleFlagsHasYOffset'] == 0 && $textRecord['styleFlagsHasXOffset'] == 0) {
                break;
            }

            if ($textRecord['styleFlagsHasFont'] != 0) {
                $textRecord['fontId'] = $this->io->collectUI16();
            }
            if ($textRecord['styleFlagsHasColor'] != 0) {
                $textRecord['textColor'] = $textVersion == 1 ? $this->collectRGB() : $this->collectRGBA();
            }
            if ($textRecord['styleFlagsHasXOffset'] != 0) {
                $textRecord['xOffset'] = $this->io->collectSI16();
            }
            if ($textRecord['styleFlagsHasYOffset'] != 0) {
                $textRecord['yOffset'] = $this->io->collectSI16();
            }
            if ($textRecord['styleFlagsHasFont'] != 0) {
                $textRecord['textHeight'] = $this->io->collectUI16();
            }
            $textRecord['glyphEntries'] = array();
            $glyphCount = $this->io->collectUI8();
            for ($i = 0; $i < $glyphCount; $i++) {
                $glyphEntry = array();
                $glyphEntry['glyphIndex'] = $this->io->collectUB($glyphBits);
                $glyphEntry['glyphAdvance'] = $this->io->collectSB($advanceBits);
                $textRecord['glyphEntries'][] = $glyphEntry;
            }
            $textRecords[] = $textRecord;
            $this->io->byteAlign();
        }
        return $textRecords;
    }

    public function collectFillStyleArray($shapeVersion) {
        $fillStyleCount = $this->io->collectUI8();
        if ($shapeVersion == 2 || $shapeVersion == 3 || $shapeVersion == 4) { //XXX shapeversion 4 not in spec
            if ($fillStyleCount == 0xff) {
                $fillStyleCount = $this->io->collectUI16(); // Extended
            }
        }
        $fillStyleArray = array();
        for ($i = 0; $i < $fillStyleCount; $i++) {
            $fillStyleArray[] = $this->collectFillStyle($shapeVersion);
        }
        return $fillStyleArray;
    }

    public function collectLineStyleArray($shapeVersion) {
        $lineStyleArray = array();
        $lineStyleCount = $this->io->collectUI8();
        if ($lineStyleCount == 0xff) {
            $lineStyleCount = $this->io->collectUI16(); // Extended
        }
        if ($shapeVersion == 1 || $shapeVersion == 2 || $shapeVersion == 3) {
            for ($i = 0; $i < $lineStyleCount; $i++) {
                $lineStyle = array();
                $lineStyle['width'] = $this->io->collectUI16();
                $lineStyle['color'] = $shapeVersion == 1 || $shapeVersion == 2 ? $this->collectRGB() : $this->collectRGBA();
                $lineStyleArray[] = $lineStyle;
            }
        } else if ($shapeVersion == 4) {
            for ($i = 0; $i < $lineStyleCount; $i++) {
                $lineStyle = array();
                $lineStyle['width'] = $this->io->collectUI16();

                $lineStyle['startCapStyle'] = $this->io->collectUB(2);
                $lineStyle['joinStyle'] = $this->io->collectUB(2);
                $lineStyle['hasFillFlag'] = $this->io->collectUB(1);
                $lineStyle['noHScaleFlag'] = $this->io->collectUB(1);
                $lineStyle['noVScaleFlag'] = $this->io->collectUB(1);
                $lineStyle['pixelHintingFlag'] = $this->io->collectUB(1);

                $this->io->collectUB(5); // Reserved, must be 0
                $lineStyle['noClose'] = $this->io->collectUB(1);
                $lineStyle['endCapStyle'] = $this->io->collectUB(2);

                if ($lineStyle['joinStyle'] == 2) {
                    $lineStyle['miterLimitFactor'] = $this->io->collectUI16();
                }
                if ($lineStyle['hasFillFlag'] == 0) {
                    $lineStyle['color'] = $this->collectRGBA();
                } else {
                    $lineStyle['fillType'] = $this->collectFillStyle($shapeVersion);
                }
                $lineStyleArray[] = $lineStyle;
            }
        } else {
            throw new \Exception(sprintf('Internal error: shapeVersion=%d', $shapeVersion));
        }
        return $lineStyleArray;
    }

    public function collectFillStyle($shapeVersion) {
        $fillStyle = array();
        $fillStyle['type'] = $this->io->collectUI8();
        switch ($fillStyle['type']) {
            case 0x00: // Solid fill
                if ($shapeVersion == 1 || $shapeVersion == 2) {
                    $fillStyle['color'] = $this->collectRGB();
                } else if ($shapeVersion == 3 || $shapeVersion == 4) { //XXX shapeVersion 4 not in spec
                    $fillStyle['color'] = $this->collectRGBA();
                } else {
                    throw new \Exception(sprintf('Internal error: shapeVersion=%d', $shapeVersion));
                }
                break;
            case 0x10: // Linear gradient fill
            case 0x12: // Radial gradient fill
            case 0x13: // Focal gradient fill
                $fillStyle['matrix'] = $this->collectMatrix();
                if ($fillStyle['type'] == 0x10 || $fillStyle['type'] == 0x12) {
                    $fillStyle['gradient'] = $this->collectGradient($shapeVersion);
                } else if ($fillStyle['type'] == 0x13) {
                    $fillStyle['focalGradient'] = $this->collectFocalGradient($shapeVersion);
                }
                break;
            case 0x40: // Repeating bitmap fill
            case 0x41: // Clipped bitmap fill
            case 0x42: // Non-smoothed repeating bitmap
            case 0x43: // Non-smoothed clipped bitmap
                $fillStyle['bitmapId'] = $this->io->collectUI16();
                $fillStyle['bitmapMatrix'] = $this->collectMatrix();
                break;
            default:
                throw new \Exception(sprintf('Internal error: fillStyleType=%d', $fillStyle['type']));
        }
        $this->io->byteAlign();
        return $fillStyle;
    }

    public function collectZoneTable($bytePosEnd) {
        $zoneRecords = array();
        while ($this->io->bytePos < $bytePosEnd) {
            $zoneData = array();
            $numZoneData = $this->io->collectUI8();
            for ($i = 0; $i < $numZoneData; $i++) {
                $alignmentCoordinate = $this->io->collectFloat16();
                $range = $this->io->collectFloat16();
                $zoneData[] = array('alignmentCoordinate' => $alignmentCoordinate, 'range' => $range);
            }
            $this->io->collectUB(6); // Reserved;
            $zoneMaskY = $this->io->collectUB(1);
            $zoneMaskX = $this->io->collectUB(1);
            $zoneRecords[] = array('zoneData' => $zoneData, 'zoneMaskY' => $zoneMaskY, 'zoneMaskX' => $zoneMaskX);
        }
        return $zoneRecords;
    }
}