<?php

namespace swf2ass\ass;

class ASSEventTime {

    /**
     * Frame duration in milliseconds
     * @var float
     */
    public float $frame_duration;

    public ASSTime $start;
    public int $start_frame;

    public ASSTime $end;
    public int $end_frame;

    public int $duration;

    public function __construct(int $startFrame, int $duration, float $frameDurationMilliseconds, int $decimalPrecision = 2){
        $this->frame_duration = $frameDurationMilliseconds;
        $this->start_frame = $startFrame;
        $this->start = new ASSTime($this->start_frame * $frameDurationMilliseconds, $decimalPrecision, true);
        $this->end_frame = $startFrame + $duration;
        $this->end = new ASSTime($this->end_frame * $frameDurationMilliseconds, $decimalPrecision, false);
        $this->duration = $duration;
    }


    public function getMillisecondsFromStartOffset(int $frameOffset) : int {
        if(($this->start_frame + $frameOffset) > $this->end_frame){
            throw new \Exception("Out of bounds: {$this->start_frame} + $frameOffset > {$this->end_frame}");
        }

        return ($this->frame_duration * $frameOffset) + $this->start->adjusted_ms_error;
    }


    public function getMillisecondsFromEndOffset(int $frameOffset) : int {
        if($frameOffset > $this->duration){
            throw new \Exception("Out of bounds: $frameOffset > {$this->duration}");
        }

        return $this->getMillisecondsFromStartOffset($this->duration - $frameOffset);
    }

    public function slice(int $frameOffset, int $frameDuration = 1) : ASSEventTime{
        if(($this->start_frame + $frameOffset + $frameDuration) > $this->end_frame){
            throw new \Exception("Out of bounds: {$this->start_frame} + $frameOffset + $frameDuration > {$this->end_frame}");
        }

        return new ASSEventTime($this->start_frame + $frameOffset, $frameDuration, $this->start->adjusted_ms_precision);
    }
}