<?php

namespace swf2ass\ass;

class ASSTime {
    const HOURS_MS = 1000 * 3600;
    const MINUTES_MS = 1000 * 60;

    public int $milliseconds_total;
    public int $milliseconds_total_adjusted;

    public int $hours;
    public int $minutes;
    public int $seconds;
    public int $milliseconds;

    public int $adjusted_ms;
    public int $adjusted_ms_precision;
    public int $adjusted_ms_error;

    public function __construct(int $ms, int $decimalPrecision = 2, bool $roundDown = true){
        if ($ms < 0) {
            throw new \LogicException("ms less than 0: $ms");
        }

        $this->milliseconds_total = $ms;
        $this->adjusted_ms_precision = $decimalPrecision;

        $this->hours = intdiv($ms, self::HOURS_MS);
        $ms -= $this->hours * self::HOURS_MS;
        $this->minutes = intdiv($ms, self::MINUTES_MS);
        $ms -= $this->minutes * self::MINUTES_MS;

        $this->seconds = intdiv($ms, 1000);
        $this->milliseconds = $ms - $this->seconds * 1000;

        $ms_adjustement = (10 * (3 - $this->adjusted_ms_precision));
        $this->adjusted_ms = intdiv($this->milliseconds, $ms_adjustement);

        $this->adjusted_ms_error = $this->milliseconds - $this->adjusted_ms * $ms_adjustement;

        if(!$roundDown and $this->adjusted_ms_error > 0){
            $this->adjusted_ms++;
            $this->adjusted_ms_error -= $ms_adjustement;
        }
        $this->milliseconds_total_adjusted = $this->milliseconds_total + $this->adjusted_ms_error;
    }

    /**
     * NOTE: libass parses milliseconds in a way that anything different than 2 decimal precisions will make it fail.
     * @return string
     */
    public function encode() : string{
        return sprintf("%01d:%02d:%02d.%0{$this->adjusted_ms_precision}d", $this->hours, $this->minutes, $this->seconds, $this->adjusted_ms);
    }
}