<?php

namespace swf2ass\ass;

use swf2ass\Constants;
use swf2ass\MatrixTransform;
use swf2ass\StyleRecord;
use swf2ass\Vector2;

class positionTag implements ASSPositioningTag {

    private Vector2 $from;
    private ?Vector2 $to;
    private int $start;
    private int $end;

    public function __construct(Vector2 $from, ?Vector2 $to, int $start, int $end) {
        $this->from = $from;
        $this->to = $to ?? $from;
        $this->start = $start;
        $this->end = $end;
    }

    public function transitionMatrixTransform(ASSLine $line, MatrixTransform $transform): ?positionTag {
        $translation = $transform->applyToVector(new Vector2(0, 0), true)->toPixel();

        $frame = $line->end - $line->start;

        $isInitialState = ($this->start === $frame and $this->end === $frame);
        $isMovingState = ($this->start < $frame and $this->end === $frame);
        $isMovedState = ($this->start < $frame and $this->end < $frame);

        if($this->to->equals($translation)){
            if($isInitialState){
                return new positionTag($this->from, $this->to,$this->start + 1, $this->end + 1);
            }elseif ($isMovingState or $isMovedState){
                return new positionTag($this->from, $this->to, $this->start, $this->end);
            }else{
                var_dump($this);
                throw new \LogicException();
            }
        }

        if($isInitialState){ //Always allow initial move
            return new positionTag($this->from, $translation, $this->start, $this->end + 1);
        }else if($isMovingState){
            $duration = $this->end - $this->start + 1;

            $direction = $this->to->sub($this->from)->normalize();
            if(abs($direction->dot($translation->normalize()) - 1) < Constants::EPSILON){ //Same direction
                $length = $this->to->sub($this->from)->divide($duration)->squaredLength();
                $length2 = $translation->sub($this->to)->squaredLength();

                //TODO: add extra leeway?
                if(abs($length - $length2) < Constants::EPSILON){ //Same length
                    return new positionTag($this->from, $translation, $this->start, $this->end + 1);
                }
            }

            return null;
        }else if($isMovedState){
            return null;
        }else{
            var_dump($this);
            throw new \LogicException();
        }
    }

    public function encode(ASSEventTime $event): string {
        $hasMoved = $this->start !== $this->end;

        $shift = $this->end - $this->start;

        if($hasMoved){
            if($shift > 1 or ASSRenderer::getSetting("smoothTransitions", false)){
                $start = $event->getMillisecondsFromStartOffset($this->start - 1);
                $end = $event->getMillisecondsFromStartOffset($this->end);
            }else{
                $start = $event->getMillisecondsFromStartOffset($this->start) - 1;
                $end = $event->getMillisecondsFromStartOffset($this->start);
            }
            return "\\move(" . $this->from->x ."," . $this->from->y ."," . $this->to->x ."," . $this->to->y .",".$start.",".$end.")";
        }
        return "\\pos(".$this->from->x ."," . $this->from->y.")";
    }

    public function equals(ASSTag $tag): bool {
        return $tag instanceof $this and $this->start === $tag->start and $this->end === $tag->end and $this->from->equals($tag->from) and $this->to->equals($tag->to);
    }

    public static function fromMatrixTransform(MatrixTransform $transform): ?positionTag {
        $translation = $transform->applyToVector(new Vector2(0, 0), true)->toPixel();
        return new positionTag($translation, $translation, 1, 1);
    }
}