<?php

namespace swf2ass\ass;

use swf2ass\Constants;
use swf2ass\MatrixTransform;
use swf2ass\StyleRecord;
use swf2ass\Vector2;

class rotationTag implements ASSPositioningTag {

    private float $x, $y, $z;

    public function __construct(float $x, float $y, float $z) {
        $this->x = $x;
        $this->y = $y;
        $this->z = $z;
    }

    public function transitionMatrixTransform(ASSLine $line, MatrixTransform $transform): ?rotationTag {
        return self::fromMatrixTransform($transform);
    }

    public function encode(ASSEventTime $event): string {
        return sprintf("\\frx%.2F\\fry%.2F\\frz%.2F", $this->x, $this->y, $this->z);
    }

    public function equals(ASSTag $tag): bool {
        return $tag instanceof $this and abs($this->x - $tag->x) <= Constants::EPSILON and abs($this->y - $tag->y) <= Constants::EPSILON and abs($this->z - $tag->z) <= Constants::EPSILON;
    }

    public static function fromMatrixTransform(MatrixTransform $transform): ?rotationTag {
        $qr = $transform->decomposeQR();

        return new rotationTag( $qr->scale->y < 0 ? 180 : 0, $qr->scale->x < 0 ? 180 : 0, $qr->rotation * (180 / M_PI));
    }
}