<?php

namespace swf2ass\math;

use MathPHP\LinearAlgebra\Matrix;
use MathPHP\LinearAlgebra\ObjectMatrix;

class RealMatrix extends ObjectMatrix{

    public function __construct($value){

        if(!$value instanceof RealMatrix){

            if($value instanceof Matrix){
                $value = $value->getMatrix();
            }
            foreach ($value as $i => $r){
                foreach ($r as $j => $v){
                    if(!($v instanceof RealNumber)){
                        $value[$i][$j] = new RealNumber($v);
                    }
                }
            }

            parent::__construct($value);
        }else{
            parent::__construct($value->A);
        }
    }

    public function get(int $i, int $j) : RealNumber {
        return parent::get($i, $j);
    }

    public function multiply($B): RealMatrix {
        return new RealMatrix(parent::multiply(new RealMatrix($B)));
    }

    public function transpose(): RealMatrix {
        return new RealMatrix(parent::transpose());
    }

    public function submatrix(int $m₁, int $n₁, int $m₂, int $n₂): RealMatrix {
        return new RealMatrix(parent::submatrix($m₁, $n₁, $m₂, $n₂));
    }

    public function isSquare(): bool {
        return $this->m === $this->n;
    }

    public function isEqual(Matrix $B, ?RealNumber $epsilon = null): bool
    {
        if (!$this->isEqualSizeAndType($B)) {
            return false;
        }

        $m = $this->m;
        $n = $this->n;
        // All elements are the same
        for ($i = 0; $i < $m; $i++) {
            for ($j = 0; $j < $n; $j++) {
                /** @var $B RealNumber[][] */
                if (!$B[$i][$j]->equalWithEpsilon($this->A[$i][$j], $epsilon)) {
                    return false;
                }
            }
        }

        return true;
    }

    public function __toString(): string {
        return \trim(\array_reduce(\array_map(
            function ($mᵢ) {
                return '[' . \implode(', ', array_map(function (RealNumber $n){return $n->toFixed(7);}, $mᵢ)) . ']';
            },
            $this->A
        ), function ($A, $mᵢ) {
            return $A . \PHP_EOL . $mᵢ;
        }));
    }

}